/* 
 * My gradient engine for gtk+.  slightly modified from the IceGradient theme,
 * to include checkboxes, scrollbars, arrows, and a few other things from the
 * Mac2 theme engine, in order to make it look like a standard toolkit.  Almost
 * no code in this theme is my own.  It is mostly the IceGradient theme, with
 * large chunks of the Mac2 theme (by Ken Lynch <kenny@lynch42.freeserve.co.uk>)
 */

/* IceGradient theme for gtk, hacked over ThinIce by Tim Gerla and Tomas gren ThinIce,
   which is based on raster's Motif theme and the Metal theme.
   Authors: Tim Gerla <timg@means.net>
            Tomas gren <stric@ing.umu.se>
            JM Perez <jose.perez@upcnet.es>
*/ 

#include "engradient_theme.h"
#include <math.h>

#define DETAIL(xx) ((detail) && (!strcmp(xx, detail)))

extern GtkStyleClass engradient_default_class;

/* internal functions */
static void         draw_hline(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x1,
			       gint x2,
			       gint y);
static void         draw_vline(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint y1,
			       gint y2,
			       gint x);
static void         draw_shadow(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);

static void         draw_polygon(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 GdkPoint * point,
				 gint npoints,
				 gint fill);
static void         draw_arrow(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       GtkArrowType arrow_type,
			       gint fill,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_diamond(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 gint x,
				 gint y,
				 gint width,
				 gint height);
static void         draw_oval(GtkStyle * style,
			      GdkWindow * window,
			      GtkStateType state_type,
			      GtkShadowType shadow_type,
			      GdkRectangle * area,
			      GtkWidget * widget,
			      gchar * detail,
			      gint x,
			      gint y,
			      gint width,
			      gint height);
static void         draw_string(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				const gchar * string);
static void         draw_box(GtkStyle * style,
			     GdkWindow * window,
			     GtkStateType state_type,
			     GtkShadowType shadow_type,
			     GdkRectangle * area,
			     GtkWidget * widget,
			     gchar * detail,
			     gint x,
			     gint y,
			     gint width,
			     gint height);
static void         draw_flat_box(GtkStyle * style,
				  GdkWindow * window,
				  GtkStateType state_type,
				  GtkShadowType shadow_type,
				  GdkRectangle * area,
				  GtkWidget * widget,
				  gchar * detail,
				  gint x,
				  gint y,
				  gint width,
				  gint height);
static void         draw_check(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_option(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);
static void         draw_cross(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_ramp(GtkStyle * style,
			      GdkWindow * window,
			      GtkStateType state_type,
			      GtkShadowType shadow_type,
			      GdkRectangle * area,
			      GtkWidget * widget,
			      gchar * detail,
			      GtkArrowType arrow_type,
			      gint x,
			      gint y,
			      gint width,
			      gint height);
static void         draw_tab(GtkStyle * style,
			     GdkWindow * window,
			     GtkStateType state_type,
			     GtkShadowType shadow_type,
			     GdkRectangle * area,
			     GtkWidget * widget,
			     gchar * detail,
			     gint x,
			     gint y,
			     gint width,
			     gint height);
static void         draw_shadow_gap(GtkStyle * style,
				    GdkWindow * window,
				    GtkStateType state_type,
				    GtkShadowType shadow_type,
				    GdkRectangle * area,
				    GtkWidget * widget,
				    gchar * detail,
				    gint x,
				    gint y,
				    gint width,
				    gint height,
				    GtkPositionType gap_side,
				    gint gap_x,
				    gint gap_width);
static void         draw_box_gap(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 gint x,
				 gint y,
				 gint width,
				 gint height,
				 GtkPositionType gap_side,
				 gint gap_x,
				 gint gap_width);
static void         draw_extension(GtkStyle * style,
				   GdkWindow * window,
				   GtkStateType state_type,
				   GtkShadowType shadow_type,
				   GdkRectangle * area,
				   GtkWidget * widget,
				   gchar * detail,
				   gint x,
				   gint y,
				   gint width,
				   gint height,
				   GtkPositionType gap_side);
static void         draw_focus(GtkStyle * style,
			       GdkWindow * window,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_slider(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height,
				GtkOrientation orientation);
static void         draw_handle(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height,
				GtkOrientation orientation);

static void draw_handlebox_slashes(GdkWindow *window,
				GdkGC *gc1,
				GdkGC *gc2,
				int width,
				int height,
				int x,
				int y);
static void engradient_tab(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);

/* internal data structs */

GtkStyleClass       engradient_default_class =
{
  2,
  2,
  draw_hline,
  draw_vline,
  draw_shadow,
  draw_polygon,
  draw_arrow,
  draw_diamond,
  draw_oval,
  draw_string,
  draw_box,
  draw_flat_box,
  draw_check,
  draw_option,
  draw_cross,
  draw_ramp,
  draw_tab,
  draw_shadow_gap,
  draw_box_gap,
  draw_extension,
  draw_focus,
  draw_slider,
  draw_handle
};



/**********************************/


static void extract_rgb_components(GdkVisual *visual, GdkColor *color)
{
  color->red = ((color->pixel >> visual->red_shift) << (16 - visual->red_prec)) & 0xFFFF;
  color->green = ((color->pixel >> visual->green_shift) << (16 - visual->green_prec)) & 0xFFFF;
  color->blue = ((color->pixel >> visual->blue_shift) << (16 - visual->blue_prec)) & 0xFFFF;
}

  
static void draw_gradient(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GdkRectangle * area,
	 gint x,
	 gint y,
	 gint width,
	 gint height,
	 gboolean horizontal)
{
  GdkVisual *visual;
  GdkGC *gc;
  GdkGCValues gc_values;
  GdkColor light, dark;
  GdkColor color;
  unsigned int i;
  
  visual = gdk_colormap_get_visual(style->colormap);
  if ((visual->type != GDK_VISUAL_DIRECT_COLOR) && (visual->type != GDK_VISUAL_TRUE_COLOR)) {
    gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x + 1, y + 1, width - 2, height - 2);
    return;
  }

  gdk_gc_get_values(style->light_gc[state_type], &gc_values);
  light.pixel = gc_values.foreground.pixel;
  gdk_gc_get_values(style->dark_gc[state_type], &gc_values);
  dark.pixel = gc_values.foreground.pixel;
  
  extract_rgb_components(visual, &light);
  extract_rgb_components(visual, &dark);

  gc = gdk_gc_new(window);
  gdk_gc_set_function(gc, GDK_COPY);
  gdk_gc_set_line_attributes(gc, 1, GDK_LINE_SOLID, gc_values.cap_style, gc_values.join_style);
  gdk_gc_set_clip_rectangle(gc, area);

  if (horizontal) {
    for(i=0; i < height; i++) {
      color.red = light.red + (float)((dark.red - light.red))*((float)i/(float)height);
      color.green = light.green + (float)((dark.green - light.green))*((float)i/(float)height);
      color.blue = light.blue + (float)((dark.blue - light.blue))*((float)i/(float)height);

      gdk_colormap_alloc_color(style->colormap, &color, FALSE, TRUE);
      gdk_gc_set_foreground(gc, &color);
      gdk_draw_line(window, gc, x, y+i, x+width, y+i);
      gdk_colormap_free_colors(style->colormap, &color, 1);
    }
  } else {
    for(i=0; i < width; i++) {
      color.red = light.red + (float)((dark.red - light.red))*((float)i/(float)width);
      color.green = light.green + (float)((dark.green - light.green))*((float)i/(float)width);
      color.blue = light.blue + (float)((dark.blue - light.blue))*((float)i/(float)width);

      gdk_colormap_alloc_color(style->colormap, &color, FALSE, TRUE);
      gdk_gc_set_foreground(gc, &color);
      gdk_draw_line(window, gc, x+i, y, x+i, y+height);
      gdk_colormap_free_colors(style->colormap, &color, 1);
    }
  }
  
  gdk_gc_destroy(gc);
}


/*
 * Explanation of the formula:
 * y|
 *  |\        /        If you use linear gradients in buttons the text is difficult to read because
 *  | \      /         the color range of the background is too wide. The idea is to make the range
 *  |  \    /          shorter in the center where the text will be. An easy way to do it would be
 *  !   -__-           to set the relative distances between steps corresponding to a something like
 * -+----------|-x                            y = (x - w/2) for x in [0,w]
 *  |          w       where w is the number of steps in the gradient.
 *                     At step w you'll have the acumulated sum of all steps and thus
 *                                   int(y, 0, w) = w/12
 * At each step you would get
 *                               int(y) = x/3 - wx/2 + wx/4
 * If you want a coefficient in [0,1]
 *                      int(y) / int(y, 0, w) = 4x/w - 6x/w + 3x/w =
 *                                            = ax + bx + cx
*/
static void draw_quadratic_gradient(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GdkRectangle * area,
	 gint x,
	 gint y,
	 gint width,
	 gint height,
	 gboolean horizontal,
	 gboolean invert)
{
  ThemeStyleData *data = style->engine_data;
  GdkVisual *visual;
  GdkGC *gc;
  GdkGCValues gc_values;
  GdkColor light, dark;
  GdkColor color;
  unsigned int i;
  float a, b, c, power, delta;
  
  if (!data || data->use_gradient == GRADIENT_OFF) {
    gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x + 1, y + 1, width - 2, height - 2);
    return;
  }

  visual = gdk_colormap_get_visual(style->colormap);
  if ((visual->type != GDK_VISUAL_DIRECT_COLOR) && (visual->type != GDK_VISUAL_TRUE_COLOR)) {
    gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x + 1, y + 1, width - 2, height - 2);
    return;
  }

  gdk_gc_get_values(style->light_gc[state_type], &gc_values);
  if (!invert) {
    light.pixel = gc_values.foreground.pixel;
  } else {
    dark.pixel = gc_values.foreground.pixel;
  }
  gdk_gc_get_values(style->dark_gc[state_type], &gc_values);
  if (!invert) {
    dark.pixel = gc_values.foreground.pixel;
  } else {
    light.pixel = gc_values.foreground.pixel;
  }
  
  extract_rgb_components(visual, &light);
  extract_rgb_components(visual, &dark);

  gc = gdk_gc_new(window);
  gdk_gc_set_function(gc, GDK_COPY);
  gdk_gc_set_line_attributes(gc, 1, GDK_LINE_SOLID, gc_values.cap_style, gc_values.join_style);
  gdk_gc_set_clip_rectangle(gc, area);

  if (horizontal) {
    a = 4.0 / (height * height * height);
    b = -6.0 / (height * height);
    c = 3.0 / height;
    for(i=0; i < height; i++) {
      power = i * i * i;
      delta = a * power; 
    	power = i * i;
    	delta += b * power;
    	power = i;
    	delta += c * power;

      color.red = light.red + (float)((dark.red - light.red))*delta;
      color.green = light.green + (float)((dark.green - light.green))*delta;
      color.blue = light.blue + (float)((dark.blue - light.blue))*delta;

      gdk_colormap_alloc_color(style->colormap, &color, FALSE, TRUE);
      gdk_gc_set_foreground(gc, &color);
      gdk_draw_line(window, gc, x, y+i, x+width, y+i);
      gdk_colormap_free_colors(style->colormap, &color, 1);
    }
  } else {
    a = 4.0 / (width * width * width);
    b = -6.0 / (width * width);
    c = 3.0 / width;
    for(i=0; i < width; i++) {
      power = i * i * i;
      delta = a * power; 
    	power = i * i;
    	delta += b * power;
    	power = i;
    	delta += c * power;

      color.red = light.red + (float)((dark.red - light.red))*delta;
      color.green = light.green + (float)((dark.green - light.green))*delta;
      color.blue = light.blue + (float)((dark.blue - light.blue))*delta;

      gdk_colormap_alloc_color(style->colormap, &color, FALSE, TRUE);
      gdk_gc_set_foreground(gc, &color);
      gdk_draw_line(window, gc, x+i, y, x+i, y+height);
      gdk_colormap_free_colors(style->colormap, &color, 1);
    }
  }
  
  gdk_gc_destroy(gc);
}


static void draw_tile_gradient(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GdkRectangle * area,
	 gint x,
	 gint y,
	 gint width,
	 gint height,
	 gboolean invert)
{
  ThemeStyleData *data = style->engine_data;
  GdkVisual *visual;
  GdkGC *gc;
  GdkGCValues gc_values;
  GdkColor light, dark;
  GdkColor color;
  unsigned int i;
  unsigned int scale;
  
  if (!data || data->use_gradient == GRADIENT_OFF) {
    gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x + 1, y + 1, width - 2, height - 2);
    return;
  }

  visual = gdk_colormap_get_visual(style->colormap);
  if ((visual->type != GDK_VISUAL_DIRECT_COLOR) && (visual->type != GDK_VISUAL_TRUE_COLOR)) {
    gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x + 1, y + 1, width - 2, height - 2);
    return;
  }

  gdk_gc_get_values(style->light_gc[state_type], &gc_values);
  if (!invert) {
    light.pixel = gc_values.foreground.pixel;
  } else {
    dark.pixel = gc_values.foreground.pixel;
  }
  gdk_gc_get_values(style->dark_gc[state_type], &gc_values);
  if (!invert) {
    dark.pixel = gc_values.foreground.pixel;
  } else {
    light.pixel = gc_values.foreground.pixel;
  }
  
  extract_rgb_components(visual, &light);
  extract_rgb_components(visual, &dark);

  gc = gdk_gc_new(window);
  gdk_gc_set_function(gc, GDK_COPY);
  gdk_gc_set_line_attributes(gc, 1, GDK_LINE_SOLID, gc_values.cap_style, gc_values.join_style);
  gdk_gc_set_clip_rectangle(gc, area);

  scale = width + height - 1;
  for(i=0; i < width; i++) {
    color.red = light.red + (float)((dark.red - light.red))*((float)i/(float)scale);
    color.green = light.green + (float)((dark.green - light.green))*((float)i/(float)scale);
    color.blue = light.blue + (float)((dark.blue - light.blue))*((float)i/(float)scale);

    gdk_colormap_alloc_color(style->colormap, &color, FALSE, TRUE);
    gdk_gc_set_foreground(gc, &color);
    gdk_draw_line(window, gc, x+i, y, x, y+i);
    gdk_colormap_free_colors(style->colormap, &color, 1);
  }
  for(; i < scale; i++) {
    color.red = light.red + (float)((dark.red - light.red))*((float)i/(float)scale);
    color.green = light.green + (float)((dark.green - light.green))*((float)i/(float)scale);
    color.blue = light.blue + (float)((dark.blue - light.blue))*((float)i/(float)scale);

    gdk_colormap_alloc_color(style->colormap, &color, FALSE, TRUE);
    gdk_gc_set_foreground(gc, &color);
    gdk_draw_line(window, gc, x+width, y+i-width, x+i-width, y+height);
    gdk_colormap_free_colors(style->colormap, &color, 1);
  }
  
  gdk_gc_destroy(gc);
}




static void
draw_hline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x1,
	   gint x2,
	   gint y)
{
  gint                thickness_light;
  gint                thickness_dark;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  thickness_light = style->klass->ythickness / 2;
  thickness_dark = style->klass->ythickness - thickness_light;

  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type], area);
  }
  for (i = 0; i < thickness_dark; i++)
  {
    gdk_draw_line(window, style->light_gc[state_type], x2 - i - 1, y + i, x2, y + i);
    gdk_draw_line(window, style->dark_gc[state_type], x1, y + i, x2 - i - 1, y + i);
  }

  y += thickness_dark;
  for (i = 0; i < thickness_light; i++)
  {
    gdk_draw_line(window, style->dark_gc[state_type], x1, y + i, x1 + thickness_light - i - 1, y + i);
    gdk_draw_line(window, style->light_gc[state_type], x1 + thickness_light - i - 1, y + i, x2, y + i);
  }
  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type], NULL);
  }
}

static void
draw_vline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint y1,
	   gint y2,
	   gint x)
{
  gint                thickness_light;
  gint                thickness_dark;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  thickness_light = style->klass->xthickness / 2;
  thickness_dark = style->klass->xthickness - thickness_light;

  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type], area);
  }
  for (i = 0; i < thickness_dark; i++)
  {
    gdk_draw_line(window, style->light_gc[state_type], x + i, y2 - i - 1, x + i, y2);
    gdk_draw_line(window, style->dark_gc[state_type], x + i, y1, x + i, y2 - i - 1);
  }

  x += thickness_dark;
  for (i = 0; i < thickness_light; i++)
  {
    gdk_draw_line(window, style->dark_gc[state_type], x + i, y1, x + i, y1 + thickness_light - i);
    gdk_draw_line(window, style->light_gc[state_type], x + i, y1 + thickness_light - i, x + i, y2);
  }
  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type], NULL);
  }
}

static void
draw_shadow(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
  GdkGC              *gc1 = NULL;
  GdkGC              *gc2 = NULL;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      return;
    case GTK_SHADOW_IN:
    case GTK_SHADOW_ETCHED_IN:
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
      break;
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
      break;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, area);
      gdk_gc_set_clip_rectangle(gc2, area);
      if ((shadow_type == GTK_SHADOW_IN) ||
	  (shadow_type == GTK_SHADOW_OUT))
	{
	  gdk_gc_set_clip_rectangle(style->black_gc, area);
	  gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
	}
    }
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      break;

    case GTK_SHADOW_ETCHED_IN:
      gdk_draw_line(window, gc2, x, y, x + width-1, y);
      gdk_draw_line(window, gc2, x, y, x, y + height-1);
      gdk_draw_line(window, gc1, x, y + height-1, x + width-1, y + height-1);
      gdk_draw_line(window, gc1, x + width-1, y, x + width-1, y + height-1);
      gdk_draw_line(window, gc1, x+1, y+1, x + width-2, y+1);
      gdk_draw_line(window, gc1, x+1, y+1, x+1, y + height-2);
      gdk_draw_line(window, gc2, x+1, y + height-2, x + width-2, y + height-2);
      gdk_draw_line(window, gc2, x + width-2, y+1, x + width-2, y + height-2);
	    break;
    case GTK_SHADOW_IN:
      gdk_draw_line(window, gc2, x, y, x + width-1, y);
      gdk_draw_line(window, gc2, x, y, x, y + height-1);
      gdk_draw_line(window, gc1, x, y + height-1, x + width-1, y + height-1);
      gdk_draw_line(window, gc1, x + width-1, y, x + width-1, y + height-1);
	    break;
    case GTK_SHADOW_ETCHED_OUT:
      gdk_draw_line(window, gc2, x, y, x + width-1, y);
      gdk_draw_line(window, gc2, x, y, x, y + height-1);
      gdk_draw_line(window, gc1, x, y + height-1, x + width-1, y + height-1);
      gdk_draw_line(window, gc1, x + width-1, y, x + width-1, y + height-1);
      gdk_draw_line(window, gc1, x+1, y+1, x + width-2, y+1);
      gdk_draw_line(window, gc1, x+1, y+1, x+1, y + height-2);
      gdk_draw_line(window, gc2, x+1, y + height-2, x + width-2, y + height-2);
      gdk_draw_line(window, gc2, x + width-2, y+1, x + width-2, y + height-2);
	    break;
    case GTK_SHADOW_OUT:
      gdk_draw_line(window, gc2, x, y, x + width-1, y);
      gdk_draw_line(window, gc2, x, y, x, y + height-1);
      gdk_draw_line(window, gc1, x, y + height-1, x + width-1, y + height-1);
      gdk_draw_line(window, gc1, x + width-1, y, x + width-1, y + height-1);
	    break;
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, NULL);
      gdk_gc_set_clip_rectangle(gc2, NULL);
      if ((shadow_type == GTK_SHADOW_IN) ||
	  (shadow_type == GTK_SHADOW_OUT))
	{
	  gdk_gc_set_clip_rectangle(style->black_gc, NULL);
	  gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	}
    }
}

static void
draw_polygon(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     GdkPoint * points,
	     gint npoints,
	     gint fill)
{
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif /* M_PI */
#ifndef M_PI_4
#define M_PI_4  0.78539816339744830962
#endif /* M_PI_4 */

  static const gdouble pi_over_4 = M_PI_4;
  static const gdouble pi_3_over_4 = M_PI_4 * 3;

  GdkGC              *gc1;
  GdkGC              *gc2;
  GdkGC              *gc3;
  GdkGC              *gc4;
  gdouble             angle;
  gint                xadjust;
  gint                yadjust;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
  g_return_if_fail(points != NULL);

  switch (shadow_type)
    {
    case GTK_SHADOW_IN:
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
      gc3 = style->light_gc[state_type];
      gc4 = style->dark_gc[state_type];
      break;
    case GTK_SHADOW_ETCHED_IN:
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
      gc3 = style->dark_gc[state_type];
      gc4 = style->light_gc[state_type];
      break;
    case GTK_SHADOW_OUT:
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
      gc3 = style->dark_gc[state_type];
      gc4 = style->light_gc[state_type];
      break;
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
      gc3 = style->light_gc[state_type];
      gc4 = style->dark_gc[state_type];
      break;
    default:
      return;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, area);
      gdk_gc_set_clip_rectangle(gc2, area);
      gdk_gc_set_clip_rectangle(gc3, area);
      gdk_gc_set_clip_rectangle(gc4, area);
    }

  if (fill)
    gdk_draw_polygon(window, style->bg_gc[state_type], TRUE, points, npoints);

  npoints--;

  for (i = 0; i < npoints; i++)
    {
      if ((points[i].x == points[i + 1].x) &&
	  (points[i].y == points[i + 1].y))
	{
	  angle = 0;
	}
      else
	{
	  angle = atan2(points[i + 1].y - points[i].y,
			points[i + 1].x - points[i].x);
	}

      if ((angle > -pi_3_over_4) && (angle < pi_over_4))
	{
	  if (angle > -pi_over_4)
	    {
	      xadjust = 0;
	      yadjust = 1;
	    }
	  else
	    {
	      xadjust = 1;
	      yadjust = 0;
	    }

	  gdk_draw_line(window, gc1,
			points[i].x - xadjust, points[i].y - yadjust,
			points[i + 1].x - xadjust, points[i + 1].y - yadjust);
	  gdk_draw_line(window, gc3,
			points[i].x, points[i].y,
			points[i + 1].x, points[i + 1].y);
	}
      else
	{
	  if ((angle < -pi_3_over_4) || (angle > pi_3_over_4))
	    {
	      xadjust = 0;
	      yadjust = 1;
	    }
	  else
	    {
	      xadjust = 1;
	      yadjust = 0;
	    }

	  gdk_draw_line(window, gc4,
			points[i].x + xadjust, points[i].y + yadjust,
			points[i + 1].x + xadjust, points[i + 1].y + yadjust);
	  gdk_draw_line(window, gc2,
			points[i].x, points[i].y,
			points[i + 1].x, points[i + 1].y);
	}
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, NULL);
      gdk_gc_set_clip_rectangle(gc2, NULL);
      gdk_gc_set_clip_rectangle(gc3, NULL);
      gdk_gc_set_clip_rectangle(gc4, NULL);
    }
}

static void
draw_handlebox_slashes(GdkWindow *window, GdkGC *gc1, GdkGC *gc2,
                  int width, int height, int x, int y)
{
    int centerx, centery, thick;
    int ax1=0,ax2=0,ay1=0,ay2=0;
    
    centerx = (width - 1)/2 + x;
    centery = (height - 1)/2 + y;
    if (width > height) {
        ax1 = -2; ax2 = 1;
    } else {
        ay1 = -2; ay2 = 1;
    }
    
    thick = ((width < height?width-1:height-1) >> 1) - 2;
    gdk_draw_line(window, gc2,
            centerx - thick + ax1, centery + thick + ay1,
            centerx + thick + ax1, centery - thick + ay1);
    gdk_draw_line(window, gc1,
            centerx - thick + ax1 + ax2, centery + thick + ay1 + ay2,
            centerx + thick + ax1 + ax2, centery - thick + ay1 + ay2);
    if (width > height) {
        ax1 = 2; /* ax2 = 1; */
    } else {
        ay1 = 2; /* ay2 = 1; */
    }
    gdk_draw_line(window, gc2,
            centerx - thick + ax1, centery + thick + ay1,
            centerx + thick + ax1, centery - thick + ay1);
    gdk_draw_line(window, gc1,
            centerx - thick + ax1 + ax2, centery + thick + ay1 + ay2,
            centerx + thick + ax1 + ax2, centery - thick + ay1 + ay2);
}

static void
draw_arrow(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   GtkArrowType arrow_type,
	   gint fill,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  GdkGC *bg_gc;
  GdkGC *outline_gc;
  gint half_width, half_height;
  gint size, offset;
  gint start, end, increment;
  GdkPoint points[3];
  GtkOrientation orientation;


  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  /* Scrollbar arrows have a button-style box drawn around them, and
   * arrow is made (even) smaller. */
  if (DETAIL ("hscrollbar") || DETAIL ("vscrollbar"))
    {
      switch (arrow_type) {
      case GTK_ARROW_UP:
      case GTK_ARROW_DOWN:
        orientation = GTK_ORIENTATION_VERTICAL;
        break;
      case GTK_ARROW_LEFT:
      case GTK_ARROW_RIGHT:
        orientation = GTK_ORIENTATION_HORIZONTAL;
        break;
      }
      draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
	  orientation == GTK_ORIENTATION_HORIZONTAL,
	  shadow_type == GTK_SHADOW_IN);
      width -= 4;
      height -= 4;
      x += 2;
      y += 2;
    }
  else if (DETAIL ("menuitem"))
    {
      width -= 2;
      height -= 2;
      x += 1;
      y += 1;
    }

  /* Give up if too small. */
  if (width < 4 || height < 4)
    return;

  /* Make everything a nice size and adjust position of arrow, if
   * necessary. */
  if (height > width)
    {
      y += (height - width) / 2;
      height = width = (width - 1) | 1;
    }
  else
    {
      x += (width - height) / 2;
      height = width = (height - 1) | 1;
    }

  /* The usual GC stuff. */
  bg_gc = style->base_gc[state_type];
  outline_gc = style->black_gc;
  if (area)
    {
      gdk_gc_set_clip_rectangle (outline_gc, area);
      gdk_gc_set_clip_rectangle (bg_gc, area);
    }

  /* Set up for draw loop. */
  switch (arrow_type)
    {
    case GTK_ARROW_RIGHT:
      increment = +1;
      start = x;
      end = x + width;
      break;
    case GTK_ARROW_LEFT:
      increment = -1;
      start = x + width - 1;
      end = x - 1;
      break;
    case GTK_ARROW_DOWN:
      increment = +1;
      start = y;
      end = y + height;
      break;;
    case GTK_ARROW_UP:
      increment = -1;
      start = y + height - 1;
      end = y - 1;
      break;
    }

  /* Draw loops. */
  if (arrow_type == GTK_ARROW_LEFT || arrow_type == GTK_ARROW_RIGHT)
    {
      gdk_draw_line (window, outline_gc, start, y,
                     start, y + height - 1);
      for (start += increment; start != end;
           start += (increment << 1), y++, height -= 2)
        {
          if (fill)
            {
              gdk_draw_line (window, bg_gc,
                             start,              y + 1,
                             start,              y + height - 2);
              gdk_draw_line (window, bg_gc,
                             start + increment,  y + 1,
                             start + increment,  y + height - 2);
            }
          gdk_draw_line (window, outline_gc,
                         start,              y,
                         start + increment,  y);
          gdk_draw_line (window, outline_gc,
                         start,              y + height - 1,
                         start + increment,  y + height - 1);
        }
      gdk_draw_point (window, outline_gc, start, y);
    }
  else
    {
      gdk_draw_line (window, outline_gc, x, start,
                     x + width - 1, start);
      for (start += increment; start != end;
           start += (increment << 1), x++, width -= 2)
        {
          if (fill)
            {
              gdk_draw_line (window, bg_gc,
                             x + 1,          start,
                             x + width - 2,  start);
              gdk_draw_line (window, bg_gc,
                             x + 1,          start + increment,
                             x + width - 2,  start + increment);
            }
          gdk_draw_line (window, outline_gc,
                         x,              start,
                         x,              start + increment);
          gdk_draw_line (window, outline_gc,
                         x + width - 1,  start,
                         x + width - 1,  start + increment);
        }
      gdk_draw_point (window, outline_gc, x, start);
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (outline_gc, NULL);
      gdk_gc_set_clip_rectangle (bg_gc, NULL);
    }

}

static void
draw_diamond(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height)
{
  gint                half_width;
  gint                half_height;
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  half_width = width / 2;
  half_height = height / 2;

  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
      gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
      gdk_gc_set_clip_rectangle(style->dark_gc[state_type], area);
      gdk_gc_set_clip_rectangle(style->black_gc, area);
    }
  switch (shadow_type)
    {
    case GTK_SHADOW_IN:
      gdk_draw_line(window, style->light_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + height - 2);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + height - 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + height - 1);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + height - 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x, y + half_height,
		    x + half_width, y + height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + height,
		    x + width, y + half_height);

      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + 2);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + 1);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x, y + half_height,
		    x + half_width, y);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y,
		    x + width, y + half_height);
      break;
    case GTK_SHADOW_OUT:
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + height - 2);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + height - 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + height - 1);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + height - 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x, y + half_height,
		    x + half_width, y + height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + height,
		    x + width, y + half_height);

      gdk_draw_line(window, style->light_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + 2);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + 1);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x, y + half_height,
		    x + half_width, y);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y,
		    x + width, y + half_height);
      break;
    default:
      break;
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle(style->dark_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle(style->black_gc, NULL);
    }
}

static void
draw_oval(GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  gchar * detail,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

static void
draw_string(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    const gchar * string)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
  if (area)
    {
      gdk_gc_set_clip_rectangle(style->white_gc, area);
      gdk_gc_set_clip_rectangle(style->fg_gc[state_type], area);
    }
  if (state_type == GTK_STATE_INSENSITIVE)
    gdk_draw_string(window, style->font, style->white_gc, x + 1, y + 1, string);
  gdk_draw_string(window, style->font, style->fg_gc[state_type], x, y, string);
  if (area)
    {
      gdk_gc_set_clip_rectangle(style->white_gc, NULL);
      gdk_gc_set_clip_rectangle(style->fg_gc[state_type], NULL);
    }
}


static void
draw_box(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
  GdkGC *light_gc, *dark_gc;
  GtkOrientation      orientation;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  light_gc = style->light_gc[state_type];
  dark_gc = style->dark_gc[state_type];

  orientation = GTK_ORIENTATION_HORIZONTAL;
  if (height > width)
    orientation = GTK_ORIENTATION_VERTICAL;

#ifdef DEBUG
     printf("%p %s %i %i state_type = %d (%d)\n", detail, detail, width, height, state_type, orientation);
#endif

  if (DETAIL("trough")) {
      if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
      }
      gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
              x, y, width, height);
      if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
      }
#if 0
      draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
        orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
#endif
      gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
              detail, x, y, width, height);
  } else if (DETAIL("slider")) {
        GdkPoint pointsh[7];
        int i, rect = 0, midlines = 1;
        ThemeStyleData *data = style->engine_data;
        pointsh[0].x = x;          pointsh[0].y = y+height-1;
        pointsh[1].x = x;          pointsh[1].y = y+6;
        pointsh[2].x = x+6;        pointsh[2].y = y;
        pointsh[3].x = x+width-1;  pointsh[3].y = y;
        pointsh[4].x = x+width-1;  pointsh[4].y = y+height-7;
        pointsh[5].x = x+width-7;  pointsh[5].y = y+height-1;
        pointsh[6].x = x;          pointsh[6].y = y+height-1;

#ifdef DEBUG
        printf("Slider... x,y=%d,%d width = %d, height = %d (%d)\n",x,y,width,height, state_type);
#endif
        if (data) {
            if (data->scrollbar_marks == MARKS_OFF) {
                midlines = 0;
            }
        }
        /* too small, use rect & no midlines */
        if ((width <= 13) && (height <= 13)) {
            midlines = 0;
        }

        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
            gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
            gdk_gc_set_clip_rectangle(style->dark_gc[state_type], area);
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_ACTIVE], area);
        }

        /* Just a plain rectangle with shadow ... */
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
            draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
              orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area,
          widget, detail, x, y, width, height);
        if (midlines) {
	  if (width > height)
	  {
	      draw_vline(style, window, state_type, area, widget, detail,
			 style->klass->ythickness,
			 height - style->klass->ythickness - 1, (width / 2) - 4);
	      draw_vline(style, window, state_type, area, widget, detail,
			 style->klass->ythickness,
			 height - style->klass->ythickness - 1, (width / 2) - 1);
	      draw_vline(style, window, state_type, area, widget, detail,
			 style->klass->ythickness,
			 height - style->klass->ythickness - 1, (width / 2) + 2);
	  }
	  else
	  {
	      draw_hline(style, window, state_type, area, widget, detail,
			 style->klass->xthickness,
			 width - style->klass->xthickness - 1, (height / 2) - 4);
	      draw_hline(style, window, state_type, area, widget, detail,
			 style->klass->xthickness,
			 width - style->klass->xthickness - 1, (height / 2) - 1);
	      draw_hline(style, window, state_type, area, widget, detail,
			 style->klass->xthickness,
			 width - style->klass->xthickness - 1, (height / 2) + 2);
	  }
        }
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
            gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
            gdk_gc_set_clip_rectangle(style->dark_gc[state_type], NULL);
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_ACTIVE], NULL);
        }
    } else if (DETAIL("buttondefault")) {
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
#if 0
            gdk_draw_rectangle(window, style->bg_gc[state_type],
                    TRUE, x + 1, y + 1, width - 2, height - 2);
#endif
            draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
              TRUE, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    } else if (DETAIL("togglebutton")) {
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
#if 0
            gdk_draw_rectangle(window, style->bg_gc[state_type],
                    TRUE, x + 1, y + 1, width - 2, height - 2);
#endif
            draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
              TRUE, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    } else if (DETAIL("button")) {
        GdkPoint points1[3]; /* dark */
        GdkPoint points2[3]; /* light */
        points1[0].x = x+2;  points1[0].y = y+2;
        points1[1].x = x+10; points1[1].y = y+2;
        points1[2].x = x+2;  points1[2].y = y+10;
        points2[0].x = x+3;  points2[0].y = y+3;
        points2[1].x = x+10; points2[1].y = y+3;
        points2[2].x = x+3;  points2[2].y = y+10;

        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
#if 0
            gdk_draw_rectangle(window, style->bg_gc[state_type],
                    TRUE, x + 1, y + 1, width - 2, height - 2);
#endif
            draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
              TRUE, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        /* Paint a triangle here instead of in "buttondefault"
           which is drawn _behind_ the current button */
        if (GTK_WIDGET_HAS_DEFAULT (widget)) {
            gdk_draw_polygon(window, dark_gc,
                    FALSE, points1, 3);
            gdk_draw_polygon(window, light_gc,
                    FALSE, points2, 3);
            gdk_draw_polygon(window, style->bg_gc[GTK_STATE_SELECTED],
                    TRUE, points2, 3);
        }
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    } else if (DETAIL("bar")) {
#if 0
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_SELECTED], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[GTK_STATE_SELECTED],
                TRUE, x + 1, y + 1, width - 2, height - 2);
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_SELECTED], NULL);
        }
#endif
        draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
          TRUE, shadow_type == GTK_SHADOW_IN);
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    } else if (DETAIL("handlebox_bin")) {
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
            }
            gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
                    x, y, width, height);
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
            }
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        /*
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
         */
    } else if (DETAIL("menubar")) {
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
            }
            gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
                    x, y, width, height);
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
            }
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
    } else if (DETAIL("menuitem")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
            orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("tab")) {
        engradient_tab(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    } else if (DETAIL("notebook")) {
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
            }
            gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
                    x, y, width, height);
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
            }
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("optionmenutab")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
            orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("hruler")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
            TRUE, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("vruler")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
            FALSE, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("spinbutton")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          if (shadow_type != GTK_SHADOW_NONE) {
            draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
              orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
          } else {
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
            }
            gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
                    x, y, width, height);
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
            }
          }            
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("paned")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          draw_tile_gradient(style, window, state_type, area, x, y, width, height,
            shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else if (DETAIL("optionmenu")) {
        if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
          draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
            orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
          detail, x, y, width, height);
    } else {
        if ((!style->bg_pixmap[state_type]) ||
                (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
            }
            gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
                    x, y, width, height);
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
            }
        } else {
            gtk_style_apply_default_background(style, window,
                    widget && !GTK_WIDGET_NO_WINDOW(widget),
                    state_type, area,
                    x, y, width, height);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    }
}



static void
draw_flat_box(GtkStyle * style,
	      GdkWindow * window,
	      GtkStateType state_type,
	      GtkShadowType shadow_type,
	      GdkRectangle * area,
	      GtkWidget * widget,
	      gchar * detail,
	      gint x,
	      gint y,
	      gint width,
	      gint height)
{
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);
    
    if ((width == -1) && (height == -1))
      gdk_window_get_size(window, &width, &height);
    else if (width == -1)
      gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
      gdk_window_get_size(window, NULL, &height);
    
    if (DETAIL("text") && (state_type != GTK_STATE_SELECTED)) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("text") && (state_type == GTK_STATE_SELECTED)) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        draw_quadratic_gradient(style, window, GTK_STATE_PRELIGHT, area, x, y, width, height,
          TRUE, FALSE);
        gtk_paint_shadow(style, window, GTK_STATE_PRELIGHT, GTK_SHADOW_OUT, area, widget,
                          detail, x, y, width, height);
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("tooltip")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        gdk_draw_rectangle(window, style->bg_gc[GTK_STATE_PRELIGHT], TRUE, x+1, y+1, width - 3, height - 3);
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
       gdk_draw_rectangle(window, style->black_gc, FALSE, x, y, width - 1, height - 1);
    } else if (DETAIL("viewportbin")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_NORMAL], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[GTK_STATE_NORMAL], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_NORMAL], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }      
    } else if (DETAIL("entry_bg")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->base_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->base_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->base_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("base")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("checkbutton")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("radiobutton")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("button")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else if (DETAIL("eventbox")) {
      if ((!style->bg_pixmap[state_type]) || (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE, x, y, width, height);
        if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
	      }
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    } else {
      if ((!style->bg_pixmap[state_type])) {
        draw_quadratic_gradient(style, window, GTK_STATE_PRELIGHT, area, x, y, width, height,
          TRUE, FALSE);
        gtk_paint_shadow(style, window, GTK_STATE_PRELIGHT, GTK_SHADOW_OUT, area, widget,
                          detail, x, y, width, height);
      } else {
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, x, y, width, height);
      }
    }
}

static void
draw_check(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  draw_tile_gradient(style, window, state_type, area, x, y, width-1, height-1,
    shadow_type == GTK_SHADOW_IN);
  if (shadow_type == GTK_SHADOW_IN)
  {
      gdk_draw_line(window, style->fg_gc[GTK_STATE_ACTIVE], x+2, y+height-4,
			x+2, y+(height/2)-3);
      gdk_draw_line(window, style->fg_gc[GTK_STATE_ACTIVE], x+3, y+height-4,
			x+3, y+(height/2)-3);
      gdk_draw_line(window, style->fg_gc[GTK_STATE_ACTIVE], x+3, y+height-4, 
			x+height-2, y);
      gdk_draw_line(window, style->fg_gc[GTK_STATE_ACTIVE], x+2, y+height-4, 
			x+height-2, y + 1);
  }
  gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
    detail, x, y, width, height);
}

/* Thanks to Evan Lawrence */
static void
draw_option(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            gchar * detail,
            gint x,
            gint y,
            gint width,
            gint height)
{
  GdkGC              *gc1;
  GdkGC              *gc2;
  GdkGC              *gc3;

    if (shadow_type == GTK_SHADOW_IN)
    {
	gc1 = style->dark_gc[state_type];		
	gc2 = style->light_gc[state_type];		
	gc3 = style->bg_gc[GTK_STATE_ACTIVE];
    }
    else
    {
	gc1 = style->light_gc[state_type];		
	gc2 = style->dark_gc[state_type];		
	gc3 = style->bg_gc[GTK_STATE_NORMAL];
    }

    if (area)
    {
	gdk_gc_set_clip_rectangle(gc1, area);
	gdk_gc_set_clip_rectangle(gc2, area);
	gdk_gc_set_clip_rectangle(gc3, area);
    }

    gdk_draw_arc(window, gc3, TRUE, x, y, width, height, 0, 360 * 64);
    gdk_draw_arc(window, style->dark_gc[state_type], FALSE, 
                 x-1, y-1, width+2, height+2, 0, 360 * 64);
    gdk_draw_arc(window, gc1, FALSE, x, y, width, height, 45 * 64, 225 * 64);
    gdk_draw_arc(window, gc2, FALSE, x, y, width, height, 225 * 64, 180 * 64);

    if (shadow_type == GTK_SHADOW_IN)
    {
        gdk_draw_arc(window, style->fg_gc[GTK_STATE_ACTIVE], TRUE, 
                     x+2, y+2, width-4, height-4, 0, 360 * 64);
        gdk_draw_arc(window, style->fg_gc[GTK_STATE_ACTIVE], FALSE, 
                     x+2, y+2, width-4, height-4, 0, 360 * 64);
    }

    if (area)
    {
	gdk_gc_set_clip_rectangle(gc1, NULL);
	gdk_gc_set_clip_rectangle(gc2, NULL);
	gdk_gc_set_clip_rectangle(gc3, NULL);
    }
}


static void
draw_cross(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

static void
draw_ramp(GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  gchar * detail,
	  GtkArrowType arrow_type,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

static void
draw_tab(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

#ifdef DEBUG
  printf("draw_tab(%d,%d,%d,%d) (detail = '%s')\n", x, y, width, height, detail?detail:"<null>");
#endif
  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		x, y, width, height);
}

static void
draw_shadow_gap(GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		gchar * detail,
		gint x,
		gint y,
		gint width,
		gint height,
		GtkPositionType gap_side,
		gint gap_x,
		gint gap_width)
{
  GdkRectangle        rect;
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_shadow(style, window, state_type, shadow_type, area, widget, detail,
		   x, y, width, height);

  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    }

  gtk_style_apply_default_background(style, window,
          widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area,
          rect.x, rect.y, rect.width, rect.height);
}

static void
draw_box_gap(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height,
	     GtkPositionType gap_side,
	     gint gap_x,
	     gint gap_width)
{
  GdkRectangle        rect;
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		x, y, width, height);

  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    }

  gtk_style_apply_default_background(style, window,
          widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, rect.x, rect.y, rect.width, rect.height);
}

static void
draw_extension(GtkStyle * style,
	       GdkWindow * window,
	       GtkStateType state_type,
	       GtkShadowType shadow_type,
	       GdkRectangle * area,
	       GtkWidget * widget,
	       gchar * detail,
	       gint x,
	       gint y,
	       gint width,
	       gint height,
	       GtkPositionType gap_side)
{
  GdkRectangle        rect;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		x, y, width, height);

  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + style->klass->xthickness;
      rect.y = y;
      rect.width = width - style->klass->xthickness * 2;
      rect.height = style->klass->ythickness;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + style->klass->xthickness;
      rect.y = y + height - style->klass->ythickness;
      rect.width = width - style->klass->xthickness * 2;
      rect.height = style->klass->ythickness;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + style->klass->ythickness;
      rect.width = style->klass->xthickness;
      rect.height = height - style->klass->ythickness * 2;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - style->klass->xthickness;
      rect.y = y + style->klass->ythickness;
      rect.width = style->klass->xthickness;
      rect.height = height - style->klass->ythickness * 2;
      break;
    }

  gtk_style_apply_default_background(style, window,
          widget && !GTK_WIDGET_NO_WINDOW(widget),
          state_type, area, rect.x, rect.y, rect.width, rect.height);
}

static void
draw_focus(GtkStyle * style,
	   GdkWindow * window,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  GdkGC *gc;
  gint8 dashes[]={1,1};
  
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1)) {
    gdk_window_get_size(window, &width, &height);
    width -= 1;
    height -= 1;
  } else if (width == -1) {
      gdk_window_get_size(window, &width, NULL);
      width -= 1;
  } else if (height == -1) {
    gdk_window_get_size(window, NULL, &height);
    height -= 1;
  }
  gc = gdk_gc_new(window);
  gdk_gc_copy(gc, style->black_gc);
  gdk_gc_set_line_attributes(gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
  gdk_gc_set_dashes(gc, 0, dashes, 2);
  if (area) {
    gdk_gc_set_clip_rectangle(gc, area);
  }
  gdk_draw_rectangle(window, gc, FALSE, x, y, width, height);
  if (area) {
    gdk_gc_set_clip_rectangle(gc, NULL);
  }
  gdk_gc_destroy(gc);
}

static void
draw_slider(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

#ifdef DEBUG
  printf("draw_slider(%s, %d, %d, %d)\n", detail, x, y, orientation);
#endif
#if 0
  gtk_draw_box(style, window, state_type, shadow_type, x, y,
	       width, height);
#endif

  draw_quadratic_gradient(style, window, state_type, area, x, y, width, height,
    orientation == GTK_ORIENTATION_HORIZONTAL, shadow_type == GTK_SHADOW_IN);
  if (orientation == GTK_ORIENTATION_HORIZONTAL) {
    draw_vline(style, window, state_type, area, widget, detail,
	       style->klass->ythickness,
	       height - style->klass->ythickness - 1, width / 2);
  } else {
    draw_hline(style, window, state_type, area, widget, detail,
	       style->klass->xthickness,
	       width - style->klass->xthickness - 1, height / 2);
	}
  gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
     detail, x, y, width, height);
}

static void
draw_handle(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
    GdkGC              *light_gc, *dark_gc;
    GdkRectangle        dest;
    ThemeStyleData     *data = style->engine_data;
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if ((width == -1) && (height == -1))
        gdk_window_get_size(window, &width, &height);
    else if (width == -1)
        gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
        gdk_window_get_size(window, NULL, &height);


    gtk_paint_box(style, window, state_type, shadow_type, area, widget,
            detail, x, y, width, height);
    draw_quadratic_gradient(style, window, state_type, area, x, y+1, width-1, height-2,
      FALSE, shadow_type == GTK_SHADOW_IN);

    /* Draw // in the box if so wanted */
    if ((data) && (data->handlebox_marks == MARKS_ON)) {
        light_gc = style->light_gc[state_type];
        dark_gc = style->dark_gc[state_type];

        dest.x = x + style->klass->xthickness;
        dest.y = y + style->klass->ythickness;
        dest.width = width - (style->klass->xthickness * 2);
        dest.height = height - (style->klass->ythickness * 2);

        gdk_gc_set_clip_rectangle(light_gc, &dest);
        gdk_gc_set_clip_rectangle(dark_gc, &dest);

        draw_handlebox_slashes(window, light_gc, dark_gc, width, height, x, y);

        gdk_gc_set_clip_rectangle(light_gc, NULL);
        gdk_gc_set_clip_rectangle(dark_gc, NULL);
    }

}



static void
engradient_tab(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
    GtkNotebook *notebook;
    GdkGC *lightgc, *darkgc;
    int orientation;

    
    notebook = GTK_NOTEBOOK(widget);
    orientation = notebook->tab_pos;
    
    lightgc = style->light_gc[state_type];
    darkgc = style->dark_gc[state_type];

    if ((!style->bg_pixmap[state_type]) ||
            (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)) {
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[state_type], TRUE,
                x, y, width, height);
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
        }
    } else {
        gtk_style_apply_default_background(style, window,
                widget && !GTK_WIDGET_NO_WINDOW(widget),
                state_type, area, x, y, width, height);
    }
    if (area) {
        gdk_gc_set_clip_rectangle(style->dark_gc[state_type], area);
        gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
    }
    switch(orientation) {
        case GTK_POS_TOP:
            gdk_draw_line(window, lightgc,
                    x, y + height - 1, x, y);
            gdk_draw_line(window, lightgc,
                    x, y, x + width - 1, y);
            gdk_draw_line(window, darkgc,
                    x + width - 1, y, x + width - 1, y + height - 1);
            break;
        case GTK_POS_BOTTOM:
            gdk_draw_line(window, lightgc,
                    x, y, x, y + height - 1);
            gdk_draw_line(window, darkgc,
                    x, y + height - 1, x + width - 1, y + height - 1);
            gdk_draw_line(window, darkgc,
                    x + width - 1, y + height - 1, x + width - 1, y);
            break;
        case GTK_POS_LEFT:
            gdk_draw_line(window, lightgc,
                    x, y + height - 1, x, y);
            gdk_draw_line(window, lightgc,
                    x, y, x + width - 1, y);
            gdk_draw_line(window, darkgc,
                    x, y + height - 1, x + width - 1, y + height - 1);
            break;
        case GTK_POS_RIGHT:
            gdk_draw_line(window, lightgc,
                    x, y, x + width - 1, y);
            gdk_draw_line(window, darkgc,
                    x + width - 1, y, x + width - 1, y + height - 1);
            gdk_draw_line(window, darkgc,
                    x, y + height - 1, x + width - 1, y + height - 1);
            break;
    }
    if (area) {
        gdk_gc_set_clip_rectangle(style->dark_gc[state_type], NULL);
        gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
    }
        /*gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
                */
}
